import numpy as np

from haloreader.exceptions import UnexpectedDataTokens
from haloreader.transformer import spectral_width_factory
from haloreader.variable import Variable

from libc.stdlib cimport atof
from libc.string cimport strlen, strspn, strtok

import cython


def read_data(data_py: bytes, ngates: cython.ulong, time_vars: list[Variable], time_range_vars: list[Variable]) -> None:
    cdef char * data_c = data_py
    cdef char * token

    # Count tokens and nprofiles
    cdef unsigned long ntokens = 0
    token = strtok(data_c, " \r\n")
    while token != NULL:
        token = strtok(NULL, " \r\n")
        ntokens += 1
    cdef unsigned long ntime_vars = len(time_vars)
    cdef unsigned long ntime_range_vars = len(time_range_vars)
    cdef unsigned long nprofiles = ntokens // (ntime_vars + ngates * ntime_range_vars)
    if (ntime_vars + ngates * ntime_range_vars) * nprofiles != ntokens:
        # Some files might have extra time_range_var column for spectral width
        nprofiles = ntokens // (ntime_vars + ngates * (ntime_range_vars+1))
        if (ntime_vars + ngates * (ntime_range_vars+1)) * nprofiles == ntokens:
            ntime_range_vars += 1
            time_range_vars.append(spectral_width_factory())
        else:
            raise UnexpectedDataTokens


    data_time = np.zeros((nprofiles,ntime_vars), dtype=np.dtype("float"))
    data_time_range = np.zeros((nprofiles,ngates,ntime_range_vars), dtype=np.dtype("float"))
    cdef double [:,:] data_time_view = data_time
    cdef double [:,:,:] data_time_range_view = data_time_range



    cdef unsigned long p, tvar, g, trvar
    token = data_c
    for p in range(nprofiles):
        for tvar in range(ntime_vars):
            data_time_view[p,tvar] = atof(token)
            token += strlen(token) + 1
            token += strspn(token, " \r\n")
        for g in range(ngates):
            for trvar in range(ntime_range_vars):
                data_time_range_view[p,g,trvar] = atof(token)
                token += strlen(token) + 1
                token += strspn(token, " \r\n")

    for i,var in enumerate(time_vars):
        var.data = data_time[:,i]
        var.dimensions = ("time",)
    for i,var in enumerate(time_range_vars):
        var.data = data_time_range[:,:,i]
        var.dimensions = ("time","range")



from io import BytesIO
from typing import Any, TypeGuard

import numpy as np


def is_str_list(val: list[Any]) -> TypeGuard[list[str]]:
    return all(isinstance(x, str) for x in val)


def is_float_list(val: list[Any]) -> TypeGuard[list[float]]:
    return all(isinstance(x, float) for x in val)


def is_int_list(val: list[Any]) -> TypeGuard[list[int]]:
    return all(isinstance(x, int) for x in val)


def is_bytesio_list(val: list[Any]) -> TypeGuard[list[BytesIO]]:
    return all(isinstance(x, BytesIO) for x in val)


def is_none_list(val: list[Any]) -> TypeGuard[list[None]]:
    return all(x is None for x in val)


def is_float_or_float_list(
    val: Any | list[Any],
) -> TypeGuard[float | list[float]]:
    if isinstance(val, list):
        return all(isinstance(x, float) for x in val)
    return isinstance(val, float)


def is_str_or_str_list(
    val: Any | list[Any],
) -> TypeGuard[str | list[str]]:
    if isinstance(val, list):
        return all(isinstance(x, str) for x in val)
    return isinstance(val, str)


def is_ndarray_list(val: list[Any]) -> TypeGuard[list[np.ndarray]]:
    return all(isinstance(x, np.ndarray) for x in val)


def is_ndarray(val: Any) -> TypeGuard[np.ndarray]:
    return isinstance(val, np.ndarray)


def is_fancy_index(val: Any) -> TypeGuard[tuple[list | np.ndarray | slice]]:
    if isinstance(val, tuple):
        return all(isinstance(x, (list, np.ndarray, slice)) for x in val)
    return False



import argparse
import datetime
import logging
from glob import glob
from pathlib import Path

import matplotlib.pyplot as plt

from haloboard.writer import Writer
from halodata.datasets import get_halo_cloudnet
from haloreader.read import read, read_bg
from haloreader.type_guards import is_ndarray

log = logging.getLogger(__name__)


def halo_reader() -> None:
    logging.basicConfig(level=logging.INFO)
    args = _haloreader_args()
    if args.subcommand == "from_cloudnet":
        _from_cloudnet(args)
    elif args.subcommand == "from_raw":
        _from_raw(args)
    else:
        raise NotImplementedError


def _from_cloudnet(args: argparse.Namespace) -> None:
    halo, halobg = get_halo_cloudnet(site=args.site, date=args.date)
    if halo is None:
        log.warning("No data from %s on %s", args.site, args.date)
        return
    if halobg is None:
        raise TypeError
    log.info("Correct background")
    halo.correct_background(halobg)
    log.info("Compute beta")
    halo.compute_beta()
    log.info("Compute noise screen")
    screen = halo.compute_noise_screen()
    log.info("Compute screened beta")
    halo.compute_beta_screened(screen)
    log.info("Compute screened doppler velocity")
    halo.compute_doppler_velocity_screened(screen)
    log.info("Convert timeunits")
    halo.convert_time_unit2cloudnet_time()
    log.info("Create netCDF")
    nc_buff = halo.to_nc()
    with open(f"halo_{args.site}_{args.date}.nc", "wb") as f:
        f.write(nc_buff)
    if args.plot:
        log.info("Create plots")
        writer = Writer()
        fig, ax = plt.subplots(nplots := 6, 1, figsize=(24, nplots * 6))
        halo.intensity_raw.plot(ax[0])
        halo.doppler_velocity.plot(ax[1])
        if halo.intensity is not None:
            halo.intensity.plot(ax[2])
        if halo.beta is not None:
            halo.beta.plot(ax[3])
        if halo.beta_screened is not None:
            halo.beta_screened.plot(ax[4])
        if halo.doppler_velocity_screened is not None:
            halo.doppler_velocity_screened.plot(ax[5])

        writer.add_figure(f"halo_{args.site}_{args.date}", fig)


def _parse_files_from_arg(path_list: list) -> list:
    allowed_wildcards = "*?"
    files = []
    for src in path_list:
        if any((c in allowed_wildcards) for c in src.name):
            wildcard_files = glob(src.name)
            if wildcard_files:
                for wildcard_file in wildcard_files:
                    files.append(Path(wildcard_file))
        else:
            files.append(src)
    return files


def _from_raw(args: argparse.Namespace) -> None:
    halo_src = [src for src in args.src if src.name.endswith(".hpl")]
    halo_src = _parse_files_from_arg(halo_src)
    bg_src = [
        src
        for src in args.src
        if src.name.startswith("Background") and src.name.endswith(".txt")
    ]
    bg_src = _parse_files_from_arg(bg_src)
    halo = read(halo_src)
    if halo is None:
        log.warning("No data")
        return

    halobg = read_bg(bg_src)
    if halobg:
        if not is_ndarray(halobg.background.data):
            raise TypeError
        if halobg.background.data.shape[0] < 300:
            log.warning(
                "Few (< 300) background profiles "
                "might lead to incorrect background correction"
            )
        halo.correct_background(halobg)
    else:
        log.warning("No background files, skipping background correction")
    nc_buff = halo.to_nc()
    with args.output.open("wb") as f:
        f.write(nc_buff)
    if args.plot:
        writer = Writer()
        fig, ax = plt.subplots(3, 1, figsize=(24, 16))
        halo.intensity_raw.plot(ax[0])
        halo.doppler_velocity.plot(ax[1])
        if halo.intensity:
            halo.intensity.plot(ax[2])
        writer.add_figure(f"{args.output.stem}", fig)


def _haloreader_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        title="subcommands", dest="subcommand", required=True
    )
    _from_cloudnet_args(
        subparsers.add_parser(
            "from_cloudnet", help="Download files directly from cloudnet."
        )
    )
    _from_raw_args(
        subparsers.add_parser(
            "from_raw", help="Read and background correct raw files into netCDF"
        )
    )

    return parser.parse_args()


def _from_cloudnet_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-p", "--plot", action="store_true")
    parser.add_argument("-s", "--site", type=str, default="warsaw")
    parser.add_argument(
        "-d",
        "--date",
        type=datetime.date.fromisoformat,
        default=datetime.date.today() - datetime.timedelta(days=1),
    )


def _from_raw_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-p", "--plot", action="store_true")
    parser.add_argument(
        "src",
        type=Path,
        nargs="*",
        default=[],
        help="Raw and background files in an arbitrary order",
    )
    parser.add_argument("-o", "--output", type=Path, required=True)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, TypeGuard

import netCDF4

from haloreader.scantype import ScanType
from haloreader.type_guards import is_str_list


@dataclass(slots=True)
class Attribute:
    name: str
    value: str | ScanType | list[str]

    def nc_write(
        self,
        nc: netCDF4.Dataset,
        nc_map: dict[str, dict] | None = None,
        nc_exclude: dict[str, set] | None = None,
    ) -> None:
        nc_exclude_attr = (
            nc_exclude.get("attributes", set()) if nc_exclude is not None else set()
        )
        if self.name in nc_exclude_attr:
            return
        nc_map_attr = nc_map.get("attributes", {}) if nc_map is not None else {}
        name = nc_map_attr.get(self.name, self.name)
        if isinstance(self.value, str):
            setattr(nc, name, self.value)
        elif isinstance(self.value, ScanType):
            setattr(nc, name, str(self.value))
        elif is_str_list(self.value):
            setattr(nc, name, "\n".join(self.value))

    @classmethod
    def is_attribute_list(cls, val: list[Any]) -> TypeGuard[list[Attribute]]:
        return all(isinstance(x, Attribute) for x in val)

    @classmethod
    def merge(cls, attributes: list[Attribute]) -> Attribute | None:
        if len(attributes) == 0:
            return None
        if len(attributes) == 1:
            return attributes[0]

        first_attr = attributes[0]
        names_eq = all(first_attr.name == a.name for a in attributes[1:])
        values_eq = all(first_attr.value == a.value for a in attributes[1:])

        if names_eq and values_eq:
            return attributes[0]
        if names_eq and first_attr.name == "filename":
            return Attribute(name=first_attr.name, value=_list_values(attributes))
        raise ValueError(f"Cannot merge attribute {attributes[0].name}")


def _list_values(attributes: list[Attribute]) -> list[str]:
    value_list = [a.value for a in attributes]
    if is_str_list(value_list):
        return value_list
    raise TypeError


def _str_or_str_list2str(val: str | list[str]) -> str:
    if isinstance(val, list):
        return "\n".join(val)
    return val



import pathlib

from matplotlib.figure import Figure

from . import DEFAULT_ROOT, IMAGE_EXT


class Writer:
    def __init__(self, root: str = DEFAULT_ROOT):
        self.root = pathlib.Path(root)
        self.root.mkdir(parents=True, exist_ok=True)

    def add_figure(self, name: str, fig: Figure) -> None:
        pth = pathlib.Path(self.root, f"{name}.{IMAGE_EXT}")
        fig.savefig(pth, bbox_inches="tight")

    def add_info(self) -> None:
        raise NotImplementedError



import logging
import pkgutil
import re
from collections import Counter
from datetime import datetime, timezone
from io import BufferedReader, BytesIO
from pathlib import Path
from typing import Sequence

import lark
import numpy as np
import numpy.typing as npt
from lark.exceptions import UnexpectedInput

from haloreader.background_reader import read_background
from haloreader.data_reader import read_data
from haloreader.exceptions import BackgroundReadError
from haloreader.halo import Halo, HaloBg
from haloreader.metadata import Metadata
from haloreader.utils import UNIX_TIME_UNIT
from haloreader.variable import Variable

from .exceptions import (
    FileEmpty,
    HeaderNotFound,
    InconsistentRangeError,
    UnexpectedDataTokens,
)
from .transformer import HeaderTransformer

log = logging.getLogger(__name__)

grammar_header = pkgutil.get_data("haloreader", "grammar_header.lark")
if not isinstance(grammar_header, bytes):
    raise FileNotFoundError("Header grammar file not found")
header_parser = lark.Lark(
    grammar_header.decode(), parser="lalr", transformer=HeaderTransformer()
)


def _read_single(src: Path | BytesIO) -> Halo:
    header_end, header_bytes = _read_header(src)
    metadata, time_vars, time_range_vars, range_func = header_parser.parse(
        header_bytes.decode()
    )
    log.info("Reading data from %s", metadata.filename.value)
    data_bytes = _read_data(src, header_end)
    if not isinstance(metadata.ngates.data, int):
        raise TypeError
    read_data(data_bytes, metadata.ngates.data, time_vars, time_range_vars)
    vars_ = {var.name: var for var in time_vars + time_range_vars}
    if not _range_consistent(vars_["range"]):
        raise InconsistentRangeError
    vars_["time"] = _decimaltime2timestamp(vars_["time"], metadata)
    vars_["range"] = range_func(vars_["range"], metadata.gate_range)
    return Halo(metadata=metadata, **vars_)


def read(src_files: Sequence[Path | BytesIO]) -> Halo | None:
    halos = []
    for src in src_files:
        try:
            halos.append(_read_single(src))
        except (
            FileEmpty,
            HeaderNotFound,
            InconsistentRangeError,
            UnicodeDecodeError,
            UnexpectedInput,
            UnexpectedDataTokens,
        ) as err:
            log.warning("Skipping file", exc_info=err)
    log.info("Merging files")
    _most_common_ngates = Counter(
        halo.metadata.ngates.data
        for halo in halos
        if isinstance(halo.metadata.ngates.data, int)
    ).most_common(1)
    most_common_ngates = _most_common_ngates[0][0] if _most_common_ngates else None
    return Halo.merge(
        [halo for halo in halos if halo.metadata.ngates.data == most_common_ngates]
    )


def _range_consistent(range_var: Variable) -> bool:
    if not isinstance(range_var.dimensions, tuple):
        raise TypeError
    if not isinstance(range_var.data, np.ndarray):
        raise TypeError
    if len(range_var.dimensions) != 2 or range_var.dimensions[1] != "range":
        return False
    expected_range = np.arange(range_var.data.shape[1], dtype=range_var.data.dtype)
    if all(np.allclose(expected_range, r) for r in range_var.data):
        return True
    return False


def read_bg(
    src_files: Sequence[Path | BytesIO], filenames: list[str] | None = None
) -> HaloBg | None:
    halobgs = []
    for src, fname in _bg_src_fname_list(src_files, filenames):
        bg_bytes = _read_background(src)
        background = read_background(bg_bytes)
        if not isinstance(background.data, np.ndarray):
            raise BackgroundReadError
        time = _bgfname2timevar(fname)
        range_ = Variable(
            name="range",
            units="index",
            dimensions=("range",),
            data=np.arange(background.data.shape[1]),
        )
        halobgs.append(HaloBg(time=time, background=background, range=range_))
    _most_common_ngates = Counter(
        bg.background.data.shape[1]
        for bg in halobgs
        if isinstance(bg.background.data, np.ndarray)
    ).most_common(1)
    most_common_ngates = _most_common_ngates[0][0] if _most_common_ngates else None
    return HaloBg.merge(
        [
            bg
            for bg in halobgs
            if isinstance(bg.background.data, np.ndarray)
            and bg.background.data.shape[1] == most_common_ngates
            and not np.all(np.isclose(bg.background.data, 0))
        ]
    )


def _bgfname2timevar(fname: str) -> Variable:
    if match_ := re.match(
        r"^Background_(\d{2})(\d{2})(\d{2})-(\d{2})(\d{2})(\d{2})\.txt$", fname
    ):
        return Variable(
            name="time",
            long_name="time of a background measurement",
            units=UNIX_TIME_UNIT,
            calendar="standard",
            dimensions=("time",),
            data=np.array(
                [
                    datetime(
                        day=int(match_.group(1)),
                        month=int(match_.group(2)),
                        year=int("20" + match_.group(3)),
                        hour=int(match_.group(4)),
                        minute=int(match_.group(5)),
                        second=int(match_.group(6)),
                        tzinfo=timezone.utc,
                    ).timestamp()
                ]
            ),
        )

    raise BackgroundReadError(f"Unexpected time format in filename: {fname}")


def _bg_src_fname_list(
    src_files: Sequence[Path | BytesIO], filenames: list[str] | None
) -> Sequence[tuple[Path | BytesIO, str]]:
    src_fname_list: list[tuple[Path | BytesIO, str]] = []
    for i, src in enumerate(src_files):
        if isinstance(src, BytesIO):
            if filenames is None or len(filenames) != len(src_files):
                raise BackgroundReadError
            src_fname_list.append((src, filenames[i]))
        else:
            src_fname_list.append((src, src.name))
    return src_fname_list


def _decimaltime2timestamp(time: Variable, metadata: Metadata) -> Variable:
    if time.long_name != "decimal time" or time.units != "hours":
        raise NotImplementedError
    if metadata.start_time.units != UNIX_TIME_UNIT:
        raise NotImplementedError
    day_in_seconds = 86400
    hour_in_seconds = 3600
    if (
        not isinstance(metadata.start_time.data, np.ndarray)
        or np.ndim(metadata.start_time.data) != 1
        or metadata.start_time.data.size != 1
    ):
        raise TypeError
    t_start = np.floor(metadata.start_time.data / day_in_seconds) * day_in_seconds
    if not isinstance(time.data, np.ndarray):
        raise TypeError
    time_ = t_start + hour_in_seconds * time.data
    i_day_changed = _find_change_of_day(0, time_)
    while i_day_changed >= 0:
        time_[i_day_changed:] += day_in_seconds
        i_day_changed = _find_change_of_day(i_day_changed, time_)
    return Variable(
        name="time",
        long_name="time",
        calendar="standard",
        data=time_,
        dimensions=("time",),
        units=UNIX_TIME_UNIT,
    )


def _find_change_of_day(start: int, time: npt.NDArray) -> int:
    half_day = 43200
    for i, (time_current, time_next) in enumerate(
        zip(time[start:-1], time[start + 1 :])
    ):
        if time_current - time_next > half_day:
            return i + 1
    return -1


def _read_header(src: Path | BytesIO) -> tuple[int, bytes]:
    if isinstance(src, Path):
        with src.open("rb") as src_buf:
            return _read_header_from_bytes(src_buf)
    else:
        return _read_header_from_bytes(src)


def _read_background(src: Path | BytesIO) -> bytes:
    if isinstance(src, Path):
        with src.open("rb") as src_buf:
            return src_buf.read()
    else:
        return src.read()


def _read_header_from_bytes(
    src: BufferedReader | BytesIO,
) -> tuple[int, bytes]:
    header_end = _find_header_end(src)
    if header_end < 0:
        src.seek(0)
        if len(src.read()) == 0:
            raise FileEmpty
        raise HeaderNotFound
    header_bytes = src.read(header_end)
    return header_end, header_bytes


def _read_data(src: Path | BytesIO, header_end: int) -> bytes:
    if isinstance(src, Path):
        with src.open("rb") as src_buf:
            src_buf.seek(header_end)
            return src_buf.read()
    else:
        src.seek(header_end)
        return src.read()


def _find_header_end(src: BufferedReader | BytesIO) -> int:
    guess = 1024
    loc_end = _try_header_end(src, guess)
    if loc_end < 0:
        loc_end = _try_header_end(src, 2 * guess)
    return loc_end


def _try_header_end(file_io: BufferedReader | BytesIO, guess: int) -> int:
    pos = file_io.tell()
    fbytes = file_io.read(guess)
    file_io.seek(pos)
    loc_sep = fbytes.find(b"****")
    if loc_sep < 0:
        return -1
    loc_end = fbytes.find(b"\r\n", loc_sep)
    if loc_end >= 0:
        loc_end += 2
    return loc_end



from __future__ import annotations

import re
from enum import Enum, auto


class ScanType(Enum):
    STARE = auto()
    STARE_OVERLAPPING = auto()
    SECTORSCAN_STEPPED = auto()
    VAD = auto()
    VAD_STEPPED = auto()
    VAD_OVERLAPPING = auto()
    USER1_STEPPED = auto()
    USER1_CSM_OVERLAPPING = auto()
    USER2_STEPPED = auto()
    USER2_CSM = auto()
    WIND_PROFILE = auto()
    WIND_PROFILE_OVERLAPPING = auto()
    RHI = auto()

    def __str__(self) -> str:
        return str(self.name)

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def from_filename(cls, filename: str) -> ScanType | None:
        match_ = re.match("^([a-zA-Z0-9]+).*", filename)
        if match_:
            scantype_str = match_.group(1).upper()
            if scantype_str in cls.__members__:
                return ScanType[scantype_str]
        return None



import numpy as np
from scipy.ndimage import uniform_filter

from haloreader.type_guards import is_ndarray
from haloreader.variable import Variable


def compute_noise_screen(
    intensity: Variable, doppler_velocity: Variable, range_: Variable
) -> Variable:
    if (
        not is_ndarray(intensity.data)
        or not is_ndarray(doppler_velocity.data)
        or not is_ndarray(range_.data)
    ):
        raise TypeError
    # kernel size and threshold values have been chosen just by
    # visually checking the output
    intensity_mean_mask = uniform_filter(intensity.data, size=(21, 3)) > 1.0025
    velocity_abs_mean_mask = (
        uniform_filter(np.abs(doppler_velocity.data), size=(21, 3)) < 2
    )
    pulse_noise = np.zeros_like(intensity.data, dtype=bool)
    pulse_noise[:, range_.data < 90] = True  # 90 meters approx 3 times pulse length
    below_one = intensity.data < 1
    return Variable(
        name="noise_screen",
        dimensions=intensity.dimensions,
        data=np.logical_not(intensity_mean_mask | velocity_abs_mean_mask)
        | pulse_noise
        | below_one,
    )



# pylint: disable=C0103
import re
from datetime import datetime, timezone
from typing import Any, Callable

import lark
import numpy as np

from .attribute import Attribute
from .metadata import Metadata
from .scantype import ScanType
from .utils import UNIX_TIME_UNIT
from .variable import Variable


class HeaderTransformer(lark.Transformer):
    def header(
        self, children: list
    ) -> tuple[
        Metadata,
        list[Variable],
        list[Variable],
        Callable[[Variable, Variable], Variable],
    ]:
        _header = {}
        for child in children:
            if isinstance(child, dict):
                _header.update(child)
            elif isinstance(child, (Attribute, Variable)):
                _header[child.name] = child
            else:
                raise TypeError
        time_vars = _header.pop("time_vars")
        time_range_vars = _header.pop("time_range_vars")
        range_func_ = _header.pop("range_func")
        metadata = Metadata(**_header)
        return metadata, time_vars, time_range_vars, range_func_

    def filename(self, children: list) -> Attribute:
        val, *_ = children
        return Attribute(name="filename", value=val)

    def system_id(self, children: list) -> Attribute:
        val, *_ = children
        return Attribute(name="system_id", value=val)

    def ngates(self, children: list) -> Variable:
        val, *_ = children
        return Variable(name="ngates", long_name="number of gates", data=val)

    def npulses(self, children: list) -> Variable:
        val, *_ = children
        return Variable(name="npulses", long_name="number of pulses", data=val)

    def nrays(self, children: list) -> Variable:
        val, *_ = children
        return Variable(name="nrays", long_name="number of rays", data=val)

    def nwaypoints(self, children: list) -> Variable:
        val, *_ = children
        return Variable(name="nwaypoints", long_name="number of waypoints", data=val)

    def scantype(self, children: list) -> Attribute:
        val, *_ = children
        return Attribute(name="scantype", value=val)

    def focus_range(self, children: list) -> Variable:
        val, *_ = children
        return Variable(name="focus_range", long_name="focus range", data=val)

    def gate_range(self, children: list) -> Variable:
        val, *_ = children
        return Variable(
            name="gate_range", long_name="length of a gate", data=val, units="m"
        )

    def gate_length(self, children: list) -> Variable:
        val, *_ = children
        return Variable(
            name="gate_length",
            long_name="number of points per gate",
            data=val,
            units="1",
        )

    def resolution(self, children: list) -> Variable:
        val, *_ = children
        return Variable(
            name="resolution",
            long_name="resolution of the doppler velocity",
            data=val,
            units="m/s",
        )

    def end_of_header_with_instrument_spectral_width(self, children: list) -> Variable:
        val, *_ = children
        return Variable(
            name="instrument_spectral_width",
            long_name="instrument spectral width",
            units="1",
            data=val,
        )

    def start_time(self, children: list) -> dict:
        time_str, *_ = children
        if match_ := re.match(
            r"(\d{4})(\d{2})(\d{2}) (\d{2}):(\d{2}):(\d{2})\.(\d{2})", time_str
        ):
            return {
                "start_time": Variable(
                    name="start_time",
                    long_name="start times of individual raw files",
                    units=UNIX_TIME_UNIT,
                    calendar="standard",
                    dimensions=("start_time",),
                    data=np.array(
                        [
                            datetime(
                                year=int(match_.group(1)),
                                month=int(match_.group(2)),
                                day=int(match_.group(3)),
                                hour=int(match_.group(4)),
                                minute=int(match_.group(5)),
                                second=int(match_.group(6)),
                                microsecond=int(match_.group(7).ljust(6, "0")),
                                tzinfo=timezone.utc,
                            ).timestamp()
                        ]
                    ),
                )
            }
        raise NotImplementedError(f"Parse for fmt {time_str} not implemented yet")

    def range_of_measurement(self, children: list) -> dict:
        func, *_ = children
        return {"range_func": func}

    def RANGE_OF_MEASUREMENT(self, _: Any) -> Callable[[Variable, Variable], Variable]:
        return range_func

    time_dimension_variables = lambda self, vars: {"time_vars": vars}

    def time_dimension_variable(self, children: list) -> Variable:
        var, *_ = children
        if not isinstance(var, Variable):
            raise TypeError
        return var

    time_range_dimension_variables = lambda self, vars: {"time_range_vars": vars}

    def time_range_dimension_variable(self, children: list) -> Variable:
        var, *_ = children
        if not isinstance(var, Variable):
            raise TypeError
        return var

    DECIMAL_TIME_H = lambda self, _: Variable(
        name="time",
        units="hours",
        dimensions=("time",),
        long_name="decimal time",
    )
    AZIMUTH_DEG = lambda self, _: Variable(
        name="azimuth",
        long_name="azimuth from north",
        units="degrees",
        dimensions=("time",),
    )
    ELEVATION_DEG = lambda self, _: Variable(
        name="elevation",
        long_name="elevation from horizontal",
        units="degrees",
        dimensions=("time",),
    )
    PITCH_DEG = lambda self, _: Variable(
        name="pitch",
        long_name="pitch offset of the instrument",
        units="degrees",
        dimensions=("time",),
    )
    ROLL_DEG = lambda self, _: Variable(
        name="roll",
        long_name="roll offset of the instrument",
        units="degrees",
        dimensions=("time",),
    )

    RANGE_GATE = lambda self, _: Variable(
        name="range", long_name="index of the gate", dimensions=("time", "range")
    )
    DOPPLER = lambda self, _: Variable(
        name="doppler_velocity",
        long_name="radial velocity (positive away from lidar)",
        units="m s-1",
        dimensions=("time", "range"),
    )
    INTENSITY = lambda self, _: Variable(
        name="intensity_raw",
        long_name="raw intensity (signal-to-noise ratio + 1)",
        units="1",
        dimensions=("time", "range"),
    )
    BETA = lambda self, _: Variable(
        name="beta_raw",
        long_name="raw beta from instrument",
        units="m-1 sr-1",
        dimensions=("time", "range"),
    )
    # spectral width variable is also defined in
    # src/haloreader/data_reader/data_reader.pyx
    SPECTRAL_WIDTH = lambda self, _: spectral_width_factory()

    def scantype_enum(self, children: list[ScanType]) -> ScanType:
        val, *_ = children
        return val

    STARE = lambda self, _: ScanType.STARE
    STARE_OVERLAPPING = lambda self, _: ScanType.STARE_OVERLAPPING
    SECTORSCAN_STEPPED = lambda self, _: ScanType.SECTORSCAN_STEPPED
    VAD = lambda self, _: ScanType.VAD
    VAD_STEPPED = lambda self, _: ScanType.VAD_STEPPED
    VAD_OVERLAPPING = lambda self, _: ScanType.VAD_OVERLAPPING
    USER1_STEPPED = lambda self, _: ScanType.USER1_STEPPED
    USER1_CSM_OVERLAPPING = lambda self, _: ScanType.USER1_CSM_OVERLAPPING
    USER2_STEPPED = lambda self, _: ScanType.USER2_STEPPED
    USER2_CSM = lambda self, _: ScanType.USER2_CSM
    WIND_PROFILE = lambda self, _: ScanType.WIND_PROFILE
    WIND_PROFILE_OVERLAPPING = lambda self, _: ScanType.WIND_PROFILE_OVERLAPPING
    RHI = lambda self, _: ScanType.RHI

    STRING = lambda self, _: _.value
    INTEGER = lambda self, _: int(_.value)
    DECIMAL = lambda self, _: float(_.value)


def spectral_width_factory() -> Variable:
    return Variable(
        name="spectral_width",
        long_name="spectral width",
        units="1",
        dimensions=("time", "range"),
    )


def range_func(range_var: Variable, gate_range: Variable) -> Variable:
    if not isinstance(range_var.data, np.ndarray) or not isinstance(
        gate_range.data, float
    ):
        raise TypeError
    range_ = (range_var.data[0, :] + 0.5) * gate_range.data
    if not isinstance(range_, np.ndarray):
        raise TypeError
    return Variable(
        name="range",
        long_name="range of measurement from the instrument",
        dimensions=("range",),
        units=gate_range.units,
        data=range_,
    )



import numpy as np
from scipy.ndimage import gaussian_filter
from scipy.signal import medfilt2d

from haloreader.exceptions import BackgroundCorrectionError
from haloreader.type_guards import is_ndarray
from haloreader.variable import Variable


def background_measurement_correction(
    time: Variable,
    intensity: Variable,
    time_bg: Variable,
    bg: Variable,
    p_amp: Variable,
) -> Variable:
    """
    References
    ----------
    A novel post-processing algorithm for Halo Doppler lidars
        authors:  Ville Vakkari, Antti J. Manninen, Ewan J. O'Connor,
            Jan H. Schween, Pieter G. van Zyl, and Eleni Marinou
        doi: https://doi.org/10.5194/amt-12-839-2019
    """
    if not is_ndarray(time.data):
        raise TypeError
    if not is_ndarray(intensity.data):
        raise TypeError
    if not is_ndarray(time_bg.data):
        raise TypeError
    if not is_ndarray(bg.data):
        raise TypeError
    if not is_ndarray(p_amp.data):
        raise TypeError
    try:
        intensity_index2bg_index = _previous_measurement_map(time.data, time_bg.data)
    except ValueError as err:
        raise BackgroundCorrectionError(
            "Cannot find matching background measurement for all profiles"
        ) from err
    relevant_bg_indeces = sorted(list(set(intensity_index2bg_index)))
    relevant_bg_indeces_map = {v: i for i, v in enumerate(relevant_bg_indeces)}
    intensity_index2relevant_bg_index = np.array(
        [relevant_bg_indeces_map[i] for i in intensity_index2bg_index]
    )
    bg_relevant = bg.take(relevant_bg_indeces)
    if not is_ndarray(bg_relevant.data):
        raise TypeError
    p_amp_denormalized = (
        bg_relevant.data.sum(axis=1)[:, np.newaxis] * p_amp.data[np.newaxis, :]
    )
    background_p_amp_removed = Variable(
        name="background_p_amp_removed",
        dimensions=bg_relevant.dimensions,
        comment="p_amp removed background",
        data=bg_relevant.data - p_amp_denormalized,
    )
    bg_relevant_fit = _linear_fit(background_p_amp_removed)
    if not is_ndarray(bg_relevant_fit.data):
        raise TypeError
    return Variable(
        name="intensity",
        comment="background measurement corrected intensity",
        dimensions=intensity.dimensions,
        units=intensity.units,
        data=intensity.data
        * bg.data[intensity_index2bg_index]
        / (
            p_amp_denormalized[intensity_index2relevant_bg_index]
            + bg_relevant_fit.data[intensity_index2relevant_bg_index]
        ),
    )


def _linear_fit(background: Variable) -> Variable:
    # pylint: disable=invalid-name
    if not isinstance(background.data, np.ndarray):
        raise TypeError
    r = 3  # ignore first three gates
    n = background.data.shape[1]
    A = np.array([np.arange(n), np.ones(n)]).T
    A_inv = np.linalg.pinv(A[r:])
    x = A_inv @ background.data.T[r:]
    bg_fit = A @ x
    return Variable(name="background_fit", dimensions=("time", "range"), data=bg_fit.T)


def _previous_measurement_map(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    # pylint: disable=invalid-name
    """
    Parameters
    ----------
    a, b
        Sorted array
    Returns
    -------
    c: array
        c[i] = j where j is largest index such that b[j] <= a[i]
    """
    i = 0
    j = 0
    n_a = len(a)
    n_b = len(b)
    c: list[int | None] = [None] * n_a
    while i < n_a and j < n_b:
        if b[j] <= a[i] and (j == n_b - 1 or a[i] < b[j + 1]):
            c[i] = j
            i += 1
        elif a[i] < b[j]:
            i += 1
        else:
            j += 1
    if None in c:
        raise ValueError
    return np.array(c, dtype=int)


def threshold_signalmask(
    intensity: Variable,
    raw_threshold: float = 1.008,
    median_filter_threshold: float = 1.002,
    gaussian_threshold: float = 0.02,
) -> np.ndarray:
    """
    mask = 1, if data is signal
    mask = 0, if data is noise
    """
    if not is_ndarray(intensity.data):
        raise TypeError
    intensity_median_normalised = (
        intensity.data / np.median(intensity.data, axis=1)[:, np.newaxis]
    )
    raw_mask = intensity_median_normalised > raw_threshold
    med_mask = (
        medfilt2d(intensity_median_normalised, kernel_size=5) > median_filter_threshold
    )
    raw_or_med_mask = raw_mask | med_mask
    gaussian = gaussian_filter(raw_or_med_mask.astype(float), sigma=8, radius=16)
    gaussian_mask = np.zeros_like(raw_or_med_mask)
    gaussian_mask[gaussian > gaussian_threshold] = True
    signalmask = raw_or_med_mask | gaussian_mask
    if not is_ndarray(signalmask):
        raise TypeError
    return signalmask


def snr_correction(intensity: Variable, signalmask: np.ndarray) -> Variable:
    # pylint: disable=invalid-name
    if not is_ndarray(intensity.data):
        raise TypeError
    _mask = signalmask.copy()
    # Fit only to noise ie where signalmask is False
    _mask[:, :3] = True  # ignore first three gates since they often containt bad data
    _range = np.arange(intensity.data.shape[1], dtype=intensity.data.dtype)
    _ones = np.ones_like(_range)
    _A = np.concatenate((_range[:, np.newaxis], _ones[:, np.newaxis]), axis=1)
    A = np.repeat(_A[np.newaxis, :, :], intensity.data.shape[0], axis=0)
    A_mask = np.repeat(_mask[:, :, np.newaxis], 2, axis=2)
    A[A_mask] = 0
    A_pinv = np.linalg.pinv(A)
    _intensity = intensity.data.copy()
    _intensity[_mask] = 0
    x = A_pinv @ _intensity[:, :, np.newaxis]
    noise_fit = np.squeeze(_A[np.newaxis, :, :] @ x, axis=2)
    intensity_corrected = intensity.data.copy()
    intensity_corrected /= noise_fit
    return Variable(
        name="intensity",
        long_name="background corrected intensity",
        units=intensity.units,
        dimensions=intensity.dimensions,
        data=intensity_corrected,
    )



class HaloException(Exception):
    pass


class FileEmpty(HaloException):
    pass


class HeaderNotFound(HaloException):
    pass


class UnexpectedDataTokens(HaloException):
    pass


class InconsistentRangeError(HaloException):
    pass


class MergeError(HaloException):
    pass


class NetCDFWriteError(HaloException):
    pass


class BackgroundReadError(HaloException):
    pass


class BackgroundCorrectionError(HaloException):
    pass



import datetime
import tempfile
from pathlib import Path

import numpy as np
import pytest
from cfchecker import cfchecks

from haloreader.exceptions import FileEmpty, UnexpectedDataTokens
from haloreader.read import _read_single, read

raw_files_pass = Path("tests/raw-files/pass/")
raw_files_xfail = Path("tests/raw-files/xfail/")


def _check_cf_conventions(nc_buf):
    f = tempfile.NamedTemporaryFile(suffix=".nc")
    f.write(nc_buf)
    cf_instance = cfchecks.CFChecker()
    res = cf_instance.checker(f.name)
    err = [val["ERROR"] for val in res["variables"].values() if len(val["ERROR"]) > 0]
    assert len(err) == 0, str(res)


def test_eriswil():
    src = raw_files_pass.joinpath("eriswil-2022-12-14-Stare_91_20221214_11.hpl")
    halo = read([src])
    time = datetime.datetime.utcfromtimestamp(halo.time.data[0]).strftime(
        "%Y-%m-%d %H:%M:%S.%f"
    )
    assert time == "2022-12-14 11:00:17.979984"
    beta_raw = halo.beta_raw.data[0, 0]
    assert np.isclose(beta_raw, 1.569249e-6)
    intensity = halo.intensity_raw.data[0, 1]
    assert np.isclose(intensity, 1.014089)
    doppler_velocity = halo.doppler_velocity.data[-1, -1]
    assert np.isclose(doppler_velocity, 16.1290)
    buf = halo.to_nc()
    _check_cf_conventions(buf)


def test_eriswil_merge():
    src_11 = raw_files_pass.joinpath("eriswil-2022-12-14-Stare_91_20221214_11.hpl")
    src_12 = raw_files_pass.joinpath("eriswil-2022-12-14-Stare_91_20221214_12.hpl")
    halo = read([src_11, src_12])
    time = datetime.datetime.utcfromtimestamp(halo.time.data[0]).strftime(
        "%Y-%m-%d %H:%M:%S.%f"
    )
    assert time == "2022-12-14 11:00:17.979984"
    beta_raw = halo.beta_raw.data[0, 0]
    assert np.isclose(beta_raw, 1.569249e-6)
    intensity = halo.intensity_raw.data[0, 1]
    assert np.isclose(intensity, 1.014089)
    doppler_velocity = halo.doppler_velocity.data[-1, -1]
    assert np.isclose(doppler_velocity, -19.1484)
    buf = halo.to_nc()
    _check_cf_conventions(buf)


def test_soverato():
    src = raw_files_pass.joinpath("soverato-2021-10-01-VAD_194_20210624_170110.hpl")
    halo = read([src])
    assert halo.time.data.shape == (2,)
    assert halo.azimuth.data.shape == (2,)
    assert halo.elevation.data.shape == (2,)
    assert halo.pitch.data.shape == (2,)
    assert halo.roll.data.shape == (2,)
    assert halo.intensity_raw.data.shape == (2, 400)
    assert halo.beta_raw.data.shape == (2, 400)
    assert halo.doppler_velocity.data.shape == (2, 400)
    buf = halo.to_nc()
    _check_cf_conventions(buf)


def test_warsaw():
    src = raw_files_pass.joinpath("warsaw-2022-12-13-Stare_213_20221213_04.hpl")
    halo = read([src])
    assert halo.time.data.shape == (2,)
    assert halo.azimuth.data.shape == (2,)
    assert halo.elevation.data.shape == (2,)
    assert halo.pitch.data.shape == (2,)
    assert halo.roll.data.shape == (2,)
    assert halo.intensity_raw.data.shape == (2, 333)
    assert halo.beta_raw.data.shape == (2, 333)
    assert halo.doppler_velocity.data.shape == (2, 333)
    buf = halo.to_nc()
    _check_cf_conventions(buf)


def test_hyytiala():
    src = raw_files_pass.joinpath("hyytiala-2023-09-13-Stare_46_20230913_23.hpl")
    halo = read([src])
    assert halo.time.data.shape == (1,)
    assert halo.azimuth.data.shape == (1,)
    assert halo.elevation.data.shape == (1,)
    assert halo.pitch is None
    assert halo.roll is None
    assert halo.intensity_raw.data.shape == (1, 320)
    assert halo.beta_raw.data.shape == (1, 320)
    assert halo.doppler_velocity.data.shape == (1, 320)
    buf = halo.to_nc()
    _check_cf_conventions(buf)


def test_xfail_warsaw():
    src = raw_files_xfail.joinpath("warsaw-2021-10-01-Stare_213_20211001_18.hpl")
    with pytest.raises(UnexpectedDataTokens):
        _read_single(src)


def test_xfail_empty():
    src = raw_files_xfail.joinpath("empty.hpl")
    with pytest.raises(FileEmpty):
        _read_single(src)



import numpy
from Cython.Build import cythonize
from setuptools import Extension, setup

if __name__ == "__main__":
    extensions = [
        Extension(
            name="haloreader.data_reader",
            sources=["src/haloreader/data_reader/data_reader.pyx"],
        ),
        Extension(
            name="haloreader.background_reader",
            sources=["src/haloreader/background_reader/background_reader.pyx"],
        ),
    ]
    setup(
        ext_modules=cythonize(
            extensions,
            language_level="3",
            annotate=True,
        ),
        include_dirs=[numpy.get_include()],
        zip_safe=False,
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, TypeGuard

import netCDF4

from haloreader.type_guards import is_none_list
from haloreader.version import __version__ as pkgversion

from .attribute import Attribute
from .variable import Variable


@dataclass(slots=True)
class Metadata:
    filename: Attribute
    system_id: Attribute
    ngates: Variable
    gate_range: Variable
    gate_length: Variable
    npulses: Variable
    scantype: Attribute
    focus_range: Variable
    start_time: Variable
    resolution: Variable
    nrays: Variable | None = None
    nwaypoints: Variable | None = None
    instrument_spectral_width: Variable | None = None
    wavelength: Variable = field(
        default_factory=lambda: Variable(
            name="wavelength",
            standard_name="radiation_wavelength",
            units="m",
            data=1.565e-6,
        )
    )
    haloreader_version: Attribute = field(
        default_factory=lambda: Attribute(name="haloreader_version", value=pkgversion)
    )
    conventions: Attribute = field(
        default_factory=lambda: Attribute(name="Conventions", value="CF-1.8")
    )

    def nc_write(
        self,
        nc: netCDF4.Dataset,
        nc_map: dict[str, dict] | None = None,
        nc_exclude: dict[str, set] | None = None,
    ) -> None:
        for attr_name in self.__dataclass_fields__.keys():
            metadata_attr = getattr(self, attr_name)
            if metadata_attr is not None:
                metadata_attr.nc_write(nc, nc_map=nc_map, nc_exclude=nc_exclude)

    @classmethod
    def merge(cls, metadata_list: list[Metadata]) -> Metadata | None:
        if len(metadata_list) == 0:
            return None
        if len(metadata_list) == 1:
            return metadata_list[0]
        metadata_attrs: dict[str, Any] = {}
        for attr_name in cls.__dataclass_fields__.keys():
            metadata_attr_list = [getattr(md, attr_name) for md in metadata_list]
            if Attribute.is_attribute_list(metadata_attr_list):
                metadata_attrs[attr_name] = Attribute.merge(metadata_attr_list)
            elif Variable.is_variable_list(metadata_attr_list):
                metadata_attrs[attr_name] = Variable.merge(metadata_attr_list)
            elif is_none_list(metadata_attr_list):
                metadata_attrs[attr_name] = None
            else:
                raise TypeError
        return Metadata(**metadata_attrs)

    @classmethod
    def is_metadata_list(cls, val: list[Any]) -> TypeGuard[list[Metadata]]:
        return all(isinstance(x, Metadata) for x in val)
